@@ -6,8 +6,8 @@
 ;; Maintainer: Bozhidar Batsov <bozhidar@batsov.dev>
 ;; URL: http://github.com/bbatsov/neocaml
 ;; Keywords: languages ocaml ml
-;; Package-Version: 20260215.2148
-;; Package-Revision: f68f3ef0c1a8
+;; Package-Version: 20260218.745
+;; Package-Revision: 8db3894f034a
 ;; Package-Requires: ((emacs "29.1"))
 
 ;; This file is not part of GNU Emacs.
@@ -56,16 +56,19 @@
   :package-version '(neocaml . "0.1.0"))
 
 (defcustom neocaml-other-file-alist
-  '(("\\.mli\\'" (".ml"))
-    ("\\.ml\\'" (".mli")))
+  '(("\\.mli\\'" (".ml" ".mll" ".mly"))
+    ("\\.ml\\'" (".mli"))
+    ("\\.mll\\'" (".mli"))
+    ("\\.mly\\'" (".mli"))
+    ("\\.eliomi\\'" (".eliom"))
+    ("\\.eliom\\'" (".eliomi")))
   "Associative list of alternate extensions to find.
 See `ff-other-file-alist' and `ff-find-other-file'."
   :type '(repeat (list regexp (choice (repeat string) function)))
   :package-version '(neocaml . "0.1.0"))
 
 (defcustom neocaml-prettify-symbols-alist
-  '(("->" . ?→)
-    ("=>" . ?⇒)
+  '(("=>" . ?⇒)
     ("<-" . ?←)
     ("<=" . ?≤)
     (">=" . ?≥)
@@ -73,13 +76,30 @@ See `ff-other-file-alist' and `ff-find-other-file'."
     ("==" . ?≡)
     ("!=" . ?≢)
     ("||" . ?∨)
-    ("&&" . ?∧)
-    ("fun" . ?λ))
-  "Prettify symbols alist used by neocaml modes."
+    ("&&" . ?∧))
+  "Prettify symbols alist used by neocaml modes.
+All replacements preserve column width."
   :type '(alist :key-type string :value-type character)
   :group 'neocaml
   :package-version '(neocaml . "0.1.0"))
 
+(defcustom neocaml-prettify-symbols-extra-alist
+  '(("fun" . ?λ)
+    ("->" . ?→)
+    ("not" . ?¬))
+  "Extra prettify symbols that may affect column alignment.
+Used when `neocaml-prettify-symbols-full' is non-nil."
+  :type '(alist :key-type string :value-type character)
+  :group 'neocaml
+  :package-version '(neocaml . "0.2.0"))
+
+(defcustom neocaml-prettify-symbols-full nil
+  "When non-nil, include `neocaml-prettify-symbols-extra-alist'.
+The extra symbols may affect column alignment."
+  :type 'boolean
+  :group 'neocaml
+  :package-version '(neocaml . "0.2.0"))
+
 (defvar neocaml--debug nil
   "Enable debugging messages and show the current node in the mode-line.
 When set to t, show indentation debug info.
@@ -87,7 +107,7 @@ When set to `font-lock', show fontification info as well.
 
 Only intended for use at development time.")
 
-(defconst neocaml-version "0.1.0")
+(defconst neocaml-version "0.2.0")
 
 (defun neocaml-version ()
   "Display the current package version in the minibuffer.
@@ -129,7 +149,7 @@ Suitable for use as the value of `treesit-language-source-alist'.")
           (treesit-install-language-grammar grammar))))))
 
 ;; adapted from tuareg-mode
-(defvar neocaml-mode-syntax-table
+(defvar neocaml-base-mode-syntax-table
   (let ((st (make-syntax-table)))
     (modify-syntax-entry ?_ "_" st)
     (modify-syntax-entry ?. "'" st)     ;Make qualified names a single symbol.
@@ -439,6 +459,9 @@ The return value is suitable for `treesit-simple-indent-rules'."
      ;; let...in: body after "in" aligns with "let" (no accumulation)
      ((parent-is "let_expression") parent-bol 0)
 
+     ;; Binding operators: and* / and+ align with let* / let+
+     ((node-is "let_and_operator") parent-bol 0)
+
      ;; Let/type/external bindings: body after = is indented
      ((parent-is "let_binding") parent-bol neocaml-indent-offset)
      ((parent-is "type_binding") parent-bol neocaml-indent-offset)
@@ -745,6 +768,8 @@ OCaml uses exclusive end-columns but Emacs expects inclusive ones."
 (defun neocaml--setup-compilation ()
   "Register OCaml compilation error regexp with compile.el."
   (require 'compile)
+  ;; OCaml uses 0-indexed character positions in error messages.
+  (setq-local compilation-first-column 0)
   (setq compilation-error-regexp-alist-alist
         (assq-delete-all 'ocaml compilation-error-regexp-alist-alist))
   (push `(ocaml
@@ -756,6 +781,68 @@ OCaml uses exclusive end-columns but Emacs expects inclusive ones."
         (delq 'ocaml compilation-error-regexp-alist))
   (push 'ocaml compilation-error-regexp-alist))
 
+;;;; _build directory awareness
+
+(defun neocaml--resolve-build-path (file)
+  "Resolve FILE out of a `_build' directory, if applicable.
+For dune-style paths like `_build/default/lib/foo.ml', strip
+`_build/<context>/' and return `<project-root>/lib/foo.ml'.
+For ocamlbuild-style `_build/lib/foo.ml', strip `_build/' only.
+Return nil if FILE is not under `_build' or the resolved file
+does not exist."
+  (when-let* ((pos (string-search "/_build/" file))
+              (root (substring file 0 pos))
+              (rest (substring file (+ pos 8))) ;; skip "/_build/"
+              (components (split-string rest "/" t)))
+    (let* (;; Dune: skip first component (context dir like "default")
+           (dune-path (when (cdr components)
+                        (concat root "/" (string-join (cdr components) "/"))))
+           ;; Ocamlbuild: keep everything after _build/
+           (ocamlbuild-path (concat root "/" rest)))
+      (cond
+       ((and dune-path (file-readable-p dune-path)) dune-path)
+       ((file-readable-p ocamlbuild-path) ocamlbuild-path)))))
+
+(defun neocaml--check-build-dir ()
+  "If the current file is under `_build/', offer to switch to the source.
+Intended for use in `find-file-hook'."
+  (when-let* ((file (buffer-file-name)))
+    (when (and (string-match-p "/_build/" file)
+               (derived-mode-p 'neocaml-base-mode))
+      (if-let* ((source (neocaml--resolve-build-path file)))
+          (when (y-or-n-p (format "This file is under _build.  Switch to %s? " source))
+            (find-alternate-file source))
+        (message "Note: this file is under _build/ (no source found)")))))
+
+;;;; Fill paragraph
+
+(defun neocaml--fill-paragraph (&optional _justify)
+  "Fill the OCaml comment at point.
+Uses tree-sitter to find comment boundaries, then narrows to the
+comment body (excluding delimiters) and fills.  Returns t if point
+was in a comment, nil otherwise to let the default handler run."
+  (let* ((node (treesit-node-at (point)))
+         (comment (treesit-parent-until
+                   node
+                   (lambda (n) (equal (treesit-node-type n) "comment"))
+                   t)))
+    (when comment
+      (let ((start (treesit-node-start comment))
+            (end (treesit-node-end comment)))
+        (save-excursion
+          (save-restriction
+            ;; Narrow to comment body: skip (* prefix and *) suffix
+            (goto-char start)
+            (when (looking-at "(\\*+[ \t]*")
+              (setq start (match-end 0)))
+            (goto-char end)
+            (when (looking-back "[ \t]*\\*+)" nil)
+              (setq end (match-beginning 0)))
+            (narrow-to-region start end)
+            (let ((fill-paragraph-function nil))
+              (fill-paragraph nil))
+            t))))))
+
 ;;;; Utility commands
 
 (defconst neocaml-report-bug-url "https://github.com/bbatsov/neocaml/issues/new"
@@ -777,17 +864,31 @@ OCaml uses exclusive end-columns but Emacs expects inclusive ones."
 
 ;;;; Major mode definitions
 
-(defvar neocaml-mode-map
+;; Mode hierarchy (following the pattern used by c-ts-mode / c++-ts-mode):
+;;
+;;   prog-mode
+;;     └─ neocaml-base-mode     — shared setup: syntax table, comments,
+;;          │                      compilation, navigation, keybindings,
+;;          │                      prettify-symbols
+;;          ├─ neocaml-mode      — .ml: ocaml grammar, ml-specific imenu
+;;          └─ neocaml-interface-mode — .mli: ocaml-interface grammar,
+;;                                       mli-specific imenu
+;;
+;; Users hook into neocaml-base-mode-hook for configuration that applies
+;; to both modes.  Language-specific setup lives in neocaml--setup-mode.
+
+(defvar neocaml-base-mode-map
   (let ((map (make-sparse-keymap)))
-    (set-keymap-parent map prog-mode-map)
     (define-key map (kbd "C-c C-a") #'ff-find-other-file)
     (define-key map (kbd "C-c 4 C-a") #'ff-find-other-file-other-window)
+    (define-key map (kbd "C-c C-c") #'compile)
     (easy-menu-define neocaml-mode-menu map "Neocaml Mode Menu"
       '("OCaml"
         ("Find..."
          ["Find Interface/Implementation" ff-find-other-file]
          ["Find Interface/Implementation in other window" ff-find-other-file-other-window])
         "--"
+        ["Compile..." compile]
         ["Cycle indent function" neocaml-cycle-indent-function]
         ["Install tree-sitter grammars" neocaml-install-grammars]
         ("Documentation"
@@ -795,15 +896,13 @@ OCaml uses exclusive end-columns but Emacs expects inclusive ones."
         "--"
         ["Report a neocaml bug" neocaml-report-bug]
         ["neocaml version" neocaml-version]))
-    map))
-
-(defvar neocaml-interface-mode-map
-  (let ((map (make-sparse-keymap)))
-    (set-keymap-parent map neocaml-mode-map)))
+    map)
+  "Keymap shared by `neocaml-mode' and `neocaml-interface-mode'.")
 
 (defun neocaml--setup-mode (language)
   "Set up tree-sitter font-lock, indentation, and navigation for LANGUAGE.
-Shared setup used by both `neocaml-mode' and `neocaml-interface-mode'."
+Called from `neocaml-mode' and `neocaml-interface-mode' to configure
+the language-specific parts of the mode."
   ;; Offer to install missing grammars
   (when-let* ((missing (seq-filter (lambda (r) (not (treesit-language-available-p (car r))))
                                    neocaml-grammar-recipes)))
@@ -822,71 +921,97 @@ Shared setup used by both `neocaml-mode' and `neocaml-interface-mode'."
       ;; show the node at point in the minibuffer
       (treesit-inspect-mode))
 
-    ;; comment settings
-    (setq-local comment-start "(* ")
-    (setq-local comment-end " *)")
-    (setq-local comment-start-skip "(\\*+[ \t]*")
-
     ;; font-lock settings
     (setq-local treesit-font-lock-settings
                 (neocaml-mode--font-lock-settings language))
 
-    ;; TODO: Make this configurable?
-    (setq-local treesit-font-lock-feature-list
-                '((comment definition)
-                  (keyword string number)
-                  (attribute builtin constant type)
-                  (operator bracket delimiter variable function)))
-
     ;; indentation
     (setq-local treesit-simple-indent-rules (neocaml--indent-rules language))
-    (setq-local indent-line-function #'treesit-indent)
 
     ;; Navigation
     (when (boundp 'treesit-thing-settings)
       (setq-local treesit-thing-settings
                   (neocaml--thing-settings language)))
-    ;; Fallback for Emacs 29 (no treesit-thing-settings)
-    (unless (boundp 'treesit-thing-settings)
-      (setq-local forward-sexp-function #'neocaml-forward-sexp))
-    (setq-local treesit-defun-type-regexp
-                (cons neocaml--defun-type-regexp
-                      #'neocaml--defun-valid-p))
-    (setq-local treesit-defun-name-function #'neocaml--defun-name)
-
-    ;; ff-find-other-file setup
-    (setq-local ff-other-file-alist neocaml-other-file-alist)
-
-    ;; Setup prettify-symbols (users enable prettify-symbols-mode via hooks)
-    (setq-local prettify-symbols-alist neocaml-prettify-symbols-alist)
 
     (treesit-major-mode-setup)))
 
+(define-derived-mode neocaml-base-mode prog-mode "OCaml"
+  "Base major mode for OCaml files, providing shared setup.
+This mode is not intended to be used directly.  Use `neocaml-mode'
+for .ml files and `neocaml-interface-mode' for .mli files."
+  :syntax-table neocaml-base-mode-syntax-table
+
+  ;; comment settings
+  (setq-local comment-start "(* ")
+  (setq-local comment-end " *)")
+  (setq-local comment-start-skip "(\\*+[ \t]*")
+  (setq-local comment-end-skip "[ \t]*\\*+)")
+
+  ;; Electric indentation on delimiters
+  (setq-local electric-indent-chars
+              (append "{}()" electric-indent-chars))
+
+  ;; Fill paragraph
+  (setq-local fill-paragraph-function #'neocaml--fill-paragraph)
+  (setq-local adaptive-fill-mode t)
+
+  ;; TODO: Make this configurable?
+  (setq-local treesit-font-lock-feature-list
+              '((comment definition)
+                (keyword string number)
+                (attribute builtin constant type)
+                (operator bracket delimiter variable function)))
+
+  (setq-local indent-line-function #'treesit-indent)
+
+  ;; Fallback for Emacs 29 (no treesit-thing-settings)
+  (unless (boundp 'treesit-thing-settings)
+    (setq-local forward-sexp-function #'neocaml-forward-sexp))
+  (setq-local treesit-defun-type-regexp
+              (cons neocaml--defun-type-regexp
+                    #'neocaml--defun-valid-p))
+  (setq-local treesit-defun-name-function #'neocaml--defun-name)
+
+  ;; ff-find-other-file setup
+  (setq-local ff-other-file-alist neocaml-other-file-alist)
+
+  ;; Setup prettify-symbols (users enable prettify-symbols-mode via hooks)
+  (setq-local prettify-symbols-alist
+              (if neocaml-prettify-symbols-full
+                  (append neocaml-prettify-symbols-alist
+                          neocaml-prettify-symbols-extra-alist)
+                neocaml-prettify-symbols-alist))
+
+  (neocaml--setup-compilation))
+
 ;;;###autoload
-(define-derived-mode neocaml-mode prog-mode "OCaml"
+(define-derived-mode neocaml-mode neocaml-base-mode "OCaml"
   "Major mode for editing OCaml code.
 
-\\{neocaml-mode-map}"
-  :syntax-table neocaml-mode-syntax-table
+\\{neocaml-base-mode-map}"
   (setq-local treesit-simple-imenu-settings neocaml--imenu-settings)
-  (neocaml--setup-mode 'ocaml)
-  (neocaml--setup-compilation))
+  (neocaml--setup-mode 'ocaml))
 
 ;;;###autoload
-(define-derived-mode neocaml-interface-mode prog-mode "OCaml[Interface]"
+(define-derived-mode neocaml-interface-mode neocaml-base-mode "OCaml[Interface]"
   "Major mode for editing OCaml interface (mli) code.
 
-\\{neocaml-interface-mode-map}"
-  :syntax-table neocaml-mode-syntax-table
+\\{neocaml-base-mode-map}"
   (setq-local treesit-simple-imenu-settings neocaml--interface-imenu-settings)
-  (neocaml--setup-mode 'ocaml-interface)
-  (neocaml--setup-compilation))
+  (neocaml--setup-mode 'ocaml-interface))
 
 ;;;###autoload
 (progn
   (add-to-list 'auto-mode-alist '("\\.ml\\'" . neocaml-mode))
   (add-to-list 'auto-mode-alist '("\\.mli\\'" . neocaml-interface-mode)))
 
+;; Hide OCaml build artifacts from find-file completion
+(dolist (ext '(".cmo" ".cmx" ".cma" ".cmxa" ".cmi" ".annot" ".cmt" ".cmti"))
+  (add-to-list 'completion-ignored-extensions ext))
+
+;; Offer to switch away from _build/ copies
+(add-hook 'find-file-hook #'neocaml--check-build-dir)
+
 ;; Eglot integration: set the language IDs that ocamllsp expects.
 ;; These symbol properties are consulted by eglot when it cannot
 ;; derive the correct language-id from the major-mode name.
